<?php

/**
 * Contains logic class and interface for the grading evaluation plugin "Comparison
 * with the weightiest assessment".
 *
 * @package    workshopeval
 * @subpackage weightiest
 * @copyright  2009 David Mudrak <david.mudrak@gmail.com>
 * @copyright  2014 Albert Gasset <albertgasset@fsfe.org>
 * @license    http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */

defined('MOODLE_INTERNAL') || die();

require_once(dirname(dirname(__FILE__)) . '/best/lib.php');

class workshop_weightiest_evaluation extends workshop_best_evaluation {

    /**
     * Returns an instance of the form to provide evaluation settings.
     *
     * @return workshop_weightiest_evaluation_settings_form
     */
    public function get_settings_form(moodle_url $actionurl=null) {

        $customdata['workshop'] = $this->workshop;
        $customdata['current'] = $this->settings;
        $attributes = array('class' => 'evalsettingsform best');

        return new workshop_weightiest_evaluation_settings_form($actionurl, $customdata, 'post', '', $attributes);
    }

    /**
     * Given a list of all assessments of a single submission, updates the grading grades in database
     *
     * @param array $assessments of stdclass (->assessmentid ->assessmentweight ->reviewerid ->gradinggrade ->submissionid ->dimensionid ->grade)
     * @param array $diminfo of stdclass (->id ->weight ->max ->min)
     * @param stdClass grading evaluation settings
     * @return void
     */
    protected function process_assessments(array $assessments, array $diminfo, stdclass $settings) {
        global $DB;

        if (empty($assessments)) {
            return;
        }

        // reindex the passed flat structure to be indexed by assessmentid
        $assessments = $this->prepare_data_from_recordset($assessments);

        // normalize the dimension grades to the interval 0 - 100
        $assessments = $this->normalize_grades($assessments, $diminfo);

        // calculate the maximum weight of assessments
        $maxweight = array_reduce($assessments, function($weight, $assessment) {
            return max($weight, $assessment->weight);
        });

        // get the assessments with maximum weight
        $weightiest = array_filter($assessments, function($assesment) use ($maxweight) {
            return $assesment->weight == $maxweight;
        });

        // for every assessment, calculate its distance from the nearest weightest assessment
        $distances = array();
        foreach ($weightiest as $referential) {
            foreach ($assessments as $asid => $assessment) {
                $d = $this->assessments_distance($assessment, $referential, $diminfo, $settings);
                if (!is_null($d) and (!isset($distances[$asid]) or $d < $distances[$asid])) {
                    $distances[$asid] = $d;
                }
            }
        }

        // calculate the grading grade
        foreach ($distances as $asid => $distance) {
            $gradinggrade = (100 - $distance);
            if ($gradinggrade < 0) {
                $gradinggrade = 0;
            }
            if ($gradinggrade > 100) {
                $gradinggrade = 100;
            }
            $grades[$asid] = grade_floatval($gradinggrade);
        }

        // if the new grading grade differs from the one stored in database, update it
        // we do not use set_field() here because we want to pass $bulk param
        foreach ($grades as $assessmentid => $grade) {
            if (grade_floats_different($grade, $assessments[$assessmentid]->gradinggrade)) {
                // the value has changed
                $record = new stdclass();
                $record->id = $assessmentid;
                $record->gradinggrade = grade_floatval($grade);
                // do not set timemodified here, it contains the timestamp of when the form was
                // saved by the peer reviewer, not when it was aggregated
                $DB->update_record('workshop_assessments', $record, true);  // bulk operations expected
            }
        }
    }

    /**
     * Measures the distance of the assessment from a referential one
     *
     * The passed data structures must contain ->dimgrades property. The referential
     * assessment is supposed to be close to the average assessment. All dimension grades are supposed to be
     * normalized to the interval 0 - 100.
     * Returned value is rounded to 4 valid decimals to prevent some rounding issues - see the unit test
     * for an example.
     *
     * @param stdClass $assessment the assessment being measured
     * @param stdClass $referential assessment
     * @param array $diminfo of stdclass(->weight ->min ->max) indexed by dimension id
     * @param stdClass $settings
     * @return float|null rounded to 4 valid decimals
     */
    protected function assessments_distance(stdclass $assessment, stdclass $referential, array $diminfo, stdclass $settings) {
        $distance = 0;
        $n = 0;
        foreach (array_keys($assessment->dimgrades) as $dimid) {
            $agrade = $assessment->dimgrades[$dimid];
            $rgrade = $referential->dimgrades[$dimid];
            $weight = $diminfo[$dimid]->weight;
            $n     += $weight;

            if (abs($agrade - $rgrade) > $settings->comparison * 10 + 0.001) {
                $distance += 100 * $weight;
            }
        }
        if ($n > 0) {
            // average distance across all dimensions
            return round($distance / $n, 4);
        } else {
            return null;
        }
    }
}

/**
 * Represents the settings form for this plugin.
 */
class workshop_weightiest_evaluation_settings_form extends workshop_evaluation_settings_form {

    /**
     * Defines specific fields for this evaluation method.
     */
    protected function definition_sub() {
        $mform = $this->_form;
        $current = $this->_customdata['current'];

        $options = array();
        for ($i = 0; $i <= 9; $i++) {
            $options[$i] = get_string('comparisonlevel' . $i, 'workshopeval_weightiest');
        }
        $mform->addElement('select', 'comparison', get_string('comparison', 'workshopeval_weightiest'), $options);
        $mform->setDefault('comparison', 5);

        $this->set_data($current);
    }
}
